#include <stdio.h>
#include <stdlib.h>
#include <string.h>

struct sos {
  int attr;
  int dtadr;
  int exadr;
  int size;
  char title[256];
  unsigned char data[65536];
};

void read_obj(FILE *, struct sos *);
void read_p6(FILE *, struct sos *);
void read_mzt(FILE *, struct sos *);
void write_obj(FILE *, struct sos *);
void write_p6(FILE *, struct sos *);
void write_mzt(FILE *, struct sos *);

int main(int argc, char *argv[])
{
  FILE *fp_in;
  FILE *fp_out;
  char outfile[256];
  struct sos sos;

  if (argc < 2) {
    printf("usage: obj2p6 <filename>\n");
    printf("usage: obj2mzt <filename>\n");
    printf("usage: p62obj <filename>\n");
    printf("usage: p62mzt <filename>\n");
    printf("usage: mzt2obj <filename>\n");
    printf("usage: mzt2p6 <filename>\n");
    exit(1);
  }

  if ((fp_in = fopen(argv[1], "rb")) == NULL) {
    printf("cannot open input file: %s\n", argv[1]);
   exit(1);
  }

  strncpy(outfile, argv[1], sizeof(outfile) - 5);
  strtok(outfile, ".");

  if (strstr(argv[0], "obj2")) {
    read_obj(fp_in, &sos);
    strncpy(sos.title, outfile, sizeof(sos.title) - 1);
    sos.title[sizeof(sos.title) - 1] = '\0';
  } else if (strstr(argv[0], "p62")) {
    read_p6(fp_in, &sos);
  } else if (strstr(argv[0], "mzt2")) {
    read_mzt(fp_in, &sos);
  }
  fclose(fp_in);

  sos.attr &= 0xff;
  sos.dtadr &= 0xffff;
  sos.exadr &= 0xffff;
  sos.size &= 0xffff;

  if (sos.size == 0) {
    printf("illegal format\n");
    exit(1);
  }

  printf("ATTR:%02Xh DTADR:%04Xh EXADR:%04Xh SIZE:%04Xh\n",
	 sos.attr, sos.dtadr, sos.exadr, sos.size);

  if (strstr(argv[0], "2obj")) {
    strcat(outfile, ".obj");
  } else if (strstr(argv[0], "2p6")) {
    strcat(outfile, ".p6");
  } else if (strstr(argv[0], "2mzt")) {
    strcat(outfile, ".mzt");
  }

  if ((fp_out = fopen(outfile, "wb")) == NULL) {
    printf("cannot open output file: %s\n", outfile);
    exit(1);
  }
  printf("%s -> %s\n", argv[1], outfile);

  if (strstr(argv[0], "2obj")) {
    write_obj(fp_out, &sos);
  } else if (strstr(argv[0], "2p6")) {
    write_p6(fp_out, &sos);
  } else if (strstr(argv[0], "2mzt")) {
    write_mzt(fp_out, &sos);
  }

  fwrite(sos.data, 1, sos.size, fp_out);
  fclose(fp_out);

  return 0;
}


void read_obj(FILE *fp_in, struct sos *sos)
{
  sos->size = 0;
  if (fscanf(fp_in, "_SOS %2X %4X %4X\n", &(sos->attr), &(sos->dtadr), &(sos->exadr)) == 3) {
    sos->size = fread(&(sos->data), 1, sizeof(sos->data), fp_in);
  }
}

void write_obj(FILE *fp_out, struct sos *sos)
{
  fprintf(fp_out, "_SOS %02X %04X %04X\n", sos->attr, sos->dtadr, sos->exadr);
}

void read_p6(FILE *fp_in, struct sos *sos)
{
  unsigned char head[16];
  sos->size = 0;
  if (fread(head, 1, 16, fp_in) == 16) {
    if (head[0] == 0xd3 && head[1] == 0xd3 && head[2] == 0xd3) {
      sos->attr = head[3];
      sos->size = head[4] + head[5] * 256;
      sos->dtadr = head[6] + head[7] * 256;
      sos->exadr =  head[8] + head[9] * 256;
      strncpy(sos->title, (char *)(head + 10), 6);
      fread(&(sos->data), 1, sos->size, fp_in);
    }
  }
}

void write_p6(FILE *fp_out, struct sos *sos)
{
  fputc(0xd3, fp_out);
  fputc(0xd3, fp_out);
  fputc(0xd3, fp_out);
  fputc(sos->attr, fp_out);
  fputc(sos->size, fp_out);
  fputc(sos->size >>8, fp_out);
  fputc(sos->dtadr, fp_out);
  fputc(sos->dtadr >>8, fp_out);
  fputc(sos->exadr, fp_out);
  fputc(sos->exadr >>8, fp_out);
  if (strlen(sos->title) < 6) {
    strcat(sos->title, "      ");
  }
  fwrite(sos->title, 6, 1, fp_out);
}

void read_mzt(FILE *fp_in, struct sos *sos)
{
  unsigned char head[128];
  sos->size = 0;
  if (fread(head, 1, 128, fp_in) == 128) {
    sos->attr = head[0];
    sscanf((char *)(head + 1), "%17s", sos->title);
    sos->size = head[18] + head[19] * 256;
    sos->dtadr = head[20] + head[21] * 256;
    sos->exadr = head[22] + head[23] * 256;
    fread(&(sos->data), 1, sos->size, fp_in);
  }
}

void write_mzt(FILE *fp_out, struct sos *sos)
{
  int i;

  fputc(sos->attr, fp_out);
  if (strlen(sos->title) < 17) {
    strcat(sos->title, "                 ");
  }
  fwrite(sos->title, 17, 1, fp_out);
  fputc(sos->size, fp_out);
  fputc(sos->size >>8, fp_out);
  fputc(sos->dtadr, fp_out);
  fputc(sos->dtadr >>8, fp_out);
  fputc(sos->exadr, fp_out);
  fputc(sos->exadr >>8, fp_out);
  for (i = 0; i < 104; i++) {
    fputc(0, fp_out);
  }
}
